"""Declarative ORM event decorators and event registration.
"""

from functools import partial
from collections import namedtuple

import sqlalchemy

from ._compat import iteritems


Event = namedtuple('Event', ['name', 'attribute', 'listener', 'kargs'])


def register(cls, dct):
    """Register events defined on a class during metaclass creation."""

    events = []

    # append class attribute defined events
    if dct.get('__events__'):
        # Events defined on __events__ can have many forms (e.g. string based,
        # list of tuples, etc). So we need to iterate over them and parse into
        # standardized Event object.
        for event_name, listeners in iteritems(dct['__events__']):
            if not isinstance(listeners, list):
                listeners = [listeners]

            for listener in listeners:
                if isinstance(listener, tuple):
                    # listener definition includes event.listen keyword args
                    listener, kargs = listener
                else:
                    kargs = {}

                if not callable(listener):
                    # assume listener is a string reference to class method
                    listener = dct[listener]

                events.append(Event(
                    event_name, kargs.pop('attribute', None), listener, kargs))

    # add events which were added via @event decorator
    for value in dct.values():
        if hasattr(value, '__event__'):
            if not isinstance(value.__event__, list):  # pragma: no cover
                value.__event__ = [value.__event__]
            events.extend(value.__event__)

    if events:
        # Reassemble events dict into consistent form using Event objects as
        # values.
        events_dict = {}
        for evt in events:
            if evt.attribute is None:
                obj = cls
            else:
                obj = getattr(cls, evt.attribute)

            if evt.name.startswith('on_'):
                event_name = evt.name.replace('on_', '', 1)
            else:
                event_name = evt.name

            sqlalchemy.event.listen(obj, event_name, evt.listener, **evt.kargs)
            events_dict.setdefault(evt.name, []).append(evt)

        dct['__events__'].update(events_dict)


#####
# Event decorators
####


def event(event_names, attribute=None, no_args_func=None, **event_kargs):
    """Generic event decorator maker which attaches metadata to function object
    so that `register()` can find the event definition.
    """
    if callable(no_args_func):
        # being called as a decorator with no arguments
        return event(event_names)(no_args_func)

    def decorator(func):
        """Function decorator that attaches an `__event__` attribute hook which
        is expected when registering a method as an event handler. See
        `register()` in this module for details on how this is implemented.
        """
        if not hasattr(func, '__event__'):
            # Set initial value to list so function can handle multiple events.
            func.__event__ = []

        # NOTE: Have to assign to a separate variable name due to global name
        # access issues.
        if not isinstance(event_names, (list, tuple)):
            _event_names = [event_names]
        else:
            _event_names = event_names

        # Attach event object to function which will be picked up in
        # `register()`.
        func.__event__ += [Event(event_name, attribute, func, event_kargs)
                           for event_name in _event_names]

        # Return function as-is since method definition should be compatible
        # with sqlalchemy.event.listen().
        return func
    return decorator


def make_attribute_event(event_names):
    """Event decorator maker for attribute events."""
    return partial(event, event_names)


def make_event(event_names):
    """Event decorator maker for mapper or instance events which don't require
    an attribute.
    """
    # bind `None` to attribute argument
    return partial(event, event_names, None)

##
# Attribute Events
# http://docs.sqlalchemy.org/en/latest/orm/events.html#attribute-events
##

on_set = make_attribute_event('set')
on_append = make_attribute_event('append')
on_remove = make_attribute_event('remove')

##
# Mapper Events
# http://docs.sqlalchemy.org/en/latest/orm/events.html#mapper-events
##

before_delete = make_event('before_delete')
before_insert = make_event('before_insert')
before_update = make_event('before_update')
before_insert_update = make_event(['before_insert', 'before_update'])

after_delete = make_event('after_delete')
after_insert = make_event('after_insert')
after_update = make_event('after_update')
after_insert_update = make_event(['after_insert', 'after_update'])

on_append_result = make_event('append_result')
on_create_instance = make_event('create_instance')
on_instrument_class = make_event('instrument_class')
before_configured = make_event('before_configured')
after_configured = make_event('after_configured')
on_mapper_configured = make_event('mapper_configured')
on_populate_instance = make_event('populate_instance')
on_translate_row = make_event('translate_row')

##
# Instance Events
# http://docs.sqlalchemy.org/en/latest/orm/events.html#instance-events
##

on_expire = make_event('expire')
on_load = make_event('load')
on_refresh = make_event('refresh')

# The following events work as intended, but they don't seem like
# candidates for supporting their definition on the model class.

# @why: Having an on-some-init event defined on the model class
# seems inefficient since whatever logic they contain should be
# handled in model.__init__() anyway.
# on_first_init = make_event('first_init')
# on_init = make_event('init')
# on_init_failure = make_event('init_failure')

# @why: Again model class would already define pickle support
# so logic should be contained there and not in a separate event handler.
# on_pickle = make_event('pickle')
# on_unpickle = make_event('unpickle')

# @why: Well, not really sure how to actually trigger this event
# so don't want to support it if it doesn't have a test.
# If someone really wants this event, then it can be enabled.
# on_resurrect = make_event('resurrect')



"""Main package API entry point.

Import core objects here.
"""

from .model import ModelBase, make_declarative_base
from .query import Query, QueryModel
from .manager import Manager, ManagerMixin
from .session import Session

from .__meta__ import (
    __title__,
    __summary__,
    __url__,
    __version__,
    __author__,
    __email__,
    __license__
)

__all__ = [
    'ModelBase',
    'make_declarative_base',
    'Query',
    'QueryModel',
    'Manager',
    'ManagerMixin',
    'Session'
]



"""Declarative base class/factory and query property support.
"""

from sqlalchemy import inspect, orm
from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta

from . import query, events
from .utils import (
    classproperty, is_sequence, has_primary_key, camelcase_to_underscore)
from ._compat import iteritems


class ModelMeta(DeclarativeMeta):
    """ModelBase's metaclass which provides:

        - Tablename autogeneration
        - Multiple database binding
        - Declarative ORM events
    """
    def __new__(mcs, name, bases, dct):
        if (not dct.get('__tablename__')
                and dct.get('__table__') is None
                and has_primary_key(dct)):
            # Set to underscore version of class name.
            dct['__tablename__'] = camelcase_to_underscore(name)

        # add bind key to table info via __table_args__
        if not dct.get('__table_args__'):
            dct['__table_args__'] = {}

        dct['__table_args__'].setdefault('info', {})
        dct['__table_args__']['info']['bind_key'] = dct.get('__bind_key__')

        # Set __events__ to expected default so that it's updatable when
        # initializing. E.g. if class definition sets __events__=None but
        # defines decorated events, then we want the final __events__ attribute
        # to reflect the registered events. If set to anything that's
        # non-empty/non-dict will lead to an error if decorated events defined.
        if not dct.get('__events__'):
            dct['__events__'] = {}

        return DeclarativeMeta.__new__(mcs, name, bases, dct)

    def __init__(cls, name, bases, dct):
        super(ModelMeta, cls).__init__(name, bases, dct)
        events.register(cls, dct)


class ModelBase(object):
    """Augmentable Base class for adding shared model properties/functions"""

    # default table args
    __table_args__ = {}

    # define a default order by when not specified by query operation
    # eg: { 'order_by': [column1, column2] }
    __mapper_args__ = {}

    # register orm event listeners
    __events__ = {}

    # query class to use for self.query
    query_class = query.QueryModel

    # An instance of `query_class`.
    # Can be used to query the database for instances of this model
    # Note: requires setting Class.query = QueryProperty(session) when session
    # available. See `make_declarative_base()` for automatic implementation.
    query = None

    def __init__(self, *args, **kargs):
        self.update(*args, **kargs)

    def __repr__(self):  # pragma: no cover
        values = ', '.join(['{0}={1}'.format(c, repr(getattr(self, c)))
                            for c in self.columns])
        return '<{0}({1})>'.format(self.__class__.__name__, values)

    def update(self, data_dict=None, strict=False, **kargs):
        """Update model with arbitrary set of data."""

        data = data_dict if isinstance(data_dict, dict) else kargs

        updatable_fields = self.strict_update_fields if strict else data.keys()
        relationships = self.relationships

        for field, value in iteritems(data):
            if hasattr(self, field) and field in updatable_fields:
                # consider v a dict if any of its elements are a dict
                if is_sequence(value):
                    is_dict = any([isinstance(val, dict) for val in value])
                else:
                    is_dict = isinstance(value, dict)

                attr = getattr(self, field)

                if (hasattr(attr, 'update')
                        and is_dict
                        and not isinstance(attr, dict)):
                    # nest calls to attr.update
                    attr.update(value)
                else:
                    if field in relationships and is_dict and not value:
                        # If v is {} and we're trying to update a relationship
                        # attribute, then we need to set to None to nullify
                        # relationship value.
                        value = None
                    setattr(self, field, value)

    @property
    def __to_dict__(self):
        """Configuration for `to_dict()`. Do any necessary preprocessing and
        return a set of string attributes which represent the fields which
        should be returned when calling `to_dict()`.

        By default this model is refreshed if it's __dict__ state is empty and
        only the ORM descriptor fields are returned.

        This is the property to override if you want to return more/less than
        the default ORM descriptor fields.

        Generally, we can usually rely on `self.__dict__` as a representation
        of model when it's just been loaded from the database. In this case,
        whatever values are present in `__dict__` are the loaded values from
        the database which include/exclude lazy attributes (columns and
        relationships).

        One issue to be aware of is that after a model has been committed (or
        expired), __dict__ will be empty. This can be worked around by calling
        `self.refresh()` which will reload the data from the database using the
        default loader strategies.

        These are the two main cases this default implementation will try to
        cover. For anything more complex it would be best to override this
        property or the `to_dict()` method itself.
        """
        if not self.descriptor_dict.keys() and orm.object_session(self):
            # if the descriptor dict keys are empty, assume we need to refresh
            self.refresh()

        return set(self.descriptor_dict.keys())

    @property
    def descriptor_dict(self):
        """Return `self.__dict__` key-filtered by `self.descriptors`"""
        return dict([(key, value)
                     for key, value in iteritems(self.__dict__)
                     if key in self.descriptors])

    def to_dict(self):
        """Return dict representation of model by filtering fields using
        `self.__to_dict__`.
        """
        data = {}
        data_fields = self.__to_dict__

        for field in data_fields:
            value = getattr(self, field)

            # Nest calls to `to_dict`. Try to find method on base value,
            # sequence values, or dict values.
            if hasattr(value, 'to_dict'):
                value = value.to_dict()
            elif is_sequence(value):
                value = [v.to_dict() if hasattr(v, 'to_dict') else v
                         for v in value]
            elif isinstance(value, dict):
                value = dict([(k, v.to_dict() if hasattr(v, 'to_dict') else v)
                              for k, v in iteritems(value)])

            data[field] = value

        return data

    def __iter__(self):
        """Implement __iter__ so model can be converted to dict via dict()."""
        return iteritems(self.to_dict())

    @property
    def strict_update_fields(self):
        """Model fields which are allowed to be updated during strict mode.
        Default is to limit to table columns. Override as needed in child
        class.
        """
        return self.columns

    ##
    # session based methods/properties
    ##

    @property
    def object_session(self):
        """Return session belonging to self"""
        return orm.object_session(self)

    @classproperty
    def session(cls):
        """Return session from query property"""
        return cls.query.session

    def flush(self, *args, **kargs):
        """Call `session.flush` on `self`"""
        self.session.flush([self], *args, **kargs)

    def save(self, *args, **kargs):
        """Call `session.add` on `self`"""
        self.session.add(self, *args, **kargs)

    def delete(self, *args, **kargs):
        """Call `session.delete` on `self`"""
        return self.session.delete(self, *args, **kargs)

    def expire(self, *args, **kargs):
        """Call `session.expire` on `self`"""
        self.session.expire(self, *args, **kargs)

    def refresh(self, *args, **kargs):
        """Call `session.refresh` on `self`"""
        self.session.refresh(self, *args, **kargs)

    def expunge(self, *args, **kargs):
        """Call `session.expunge` on `self`"""
        self.session.expunge(self, *args, **kargs)

    @classmethod
    def get(cls, *args, **kargs):
        """Proxy to query.get()"""
        return cls.query.get(*args, **kargs)

    @classmethod
    def get_by(cls, data_dict=None, **kargs):
        """Return first instance filtered by values using query.filter_by()"""
        data = data_dict if isinstance(data_dict, dict) else kargs
        return cls.query.filter_by(**data).first()

    ##
    # inspect based methods/properties
    ##

    @classproperty
    def primary_key(cls):
        """Return primary key as either single column (one primary key) or
        tuple otherwise.
        """
        primary = inspect(cls).primary_key

        if len(primary) == 1:
            primary = primary[0]

        return primary

    @classproperty
    def primary_keys(cls):
        """Return primary keys as tuple."""
        return inspect(cls).primary_key

    @classproperty
    def attrs(cls):
        """Return ORM attributes"""
        return inspect(cls).attrs.keys()

    @classproperty
    def descriptors(cls):
        """Return all ORM descriptors"""
        return [descr for descr in inspect(cls).all_orm_descriptors.keys()
                if not descr.startswith('__')]

    @classproperty
    def relationships(cls):
        """Return ORM relationships"""
        return inspect(cls).relationships.keys()

    @classproperty
    def column_attrs(cls):
        """Return table columns as list of class attributes at the class level.
        """
        return inspect(cls).column_attrs

    @classproperty
    def columns(cls):
        """Return table columns"""
        return inspect(cls).columns.keys()


def make_declarative_base(session=None, Model=None, Base=None):
    """Factory function for either creating a new declarative base class or
    extending a previously defined one.
    """
    if Model is None:
        Base = Base or ModelBase
        Model = declarative_base(
            cls=Base, constructor=Base.__init__, metaclass=ModelMeta)

    extend_declarative_base(Model, session)

    return Model


def extend_declarative_base(Model, session=None):
    """Extend a declarative base class with additional properties.

    - Extend `Model` with query property accessor
    """
    # Attach query attribute to Model if `session` object passed in.
    if session:
        Model.query = query.QueryProperty(session)



"""Custom ORM column types.
"""

import re

from sqlalchemy.types import SchemaType, TypeDecorator, Enum

from ._compat import with_metaclass


##
# Declarative Enum type
# Adapted from the Enum Recipe:
# http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/
##

class EnumSymbol(object):
    """Define a fixed symbol tied to a parent class."""

    def __init__(self, cls_, name, value, description):
        self.cls_ = cls_
        self.name = name
        self.value = value
        self.description = description

    def __reduce__(self):
        """Allow unpickling to return the symbol linked to the DeclarativeEnum
        class.
        """
        return getattr, (self.cls_, self.name)

    def __iter__(self):
        return iter([self.value, self.description])

    def __repr__(self):
        return "<%s>" % self.name

    def __str__(self):
        return self.name

    def to_dict(self):
        """Represent symbol as dict."""
        return {'value': self.value, 'description': self.description}


class EnumMeta(type):
    """Generate new DeclarativeEnum classes."""

    def __init__(cls, classname, bases, dict_):
        cls._reg = reg = cls._reg.copy()
        for k, v in dict_.items():
            if isinstance(v, tuple):
                sym = reg[v[0]] = EnumSymbol(cls, k, *v)
                setattr(cls, k, sym)
        type.__init__(cls, classname, bases, dict_)

    def __iter__(cls):
        return iter(cls._reg.values())


class DeclarativeEnumType(SchemaType, TypeDecorator):
    """Column type usable in table column definitions."""

    def __init__(self, enum):
        self.enum = enum
        self.impl = Enum(
            *enum.values(),
            name="ck%s" % re.sub('([A-Z])',
                                 lambda m: "_" + m.group(1).lower(),
                                 enum.__name__)
        )

    def _set_table(self, table, column):
        self.impl._set_table(table, column)

    def copy(self):
        return DeclarativeEnumType(self.enum)

    def process_bind_param(self, value, dialect):
        if value is None:  # pragma: no cover
            return None
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:  # pragma: no cover
            return None
        return self.enum.from_string(value.strip())


class DeclarativeEnum(with_metaclass(EnumMeta, object)):
    """Declarative enumeration."""

    _reg = {}

    @classmethod
    def from_string(cls, string):
        """Get enum value from string"""
        try:
            return cls._reg[string]
        except KeyError:
            raise ValueError(
                "Invalid value for %r: %r" % (cls.__name__, string))

    @classmethod
    def values(cls):
        """Return enumeration values."""
        return cls._reg.keys()

    @classmethod
    def db_type(cls):
        """Provide database column type for use in table column definitions."""
        return DeclarativeEnumType(cls)



"""Generic utility functions used in package.
"""

import re
from collections import Iterable

from sqlalchemy import Column, and_

from ._compat import string_types, iteritems


class classproperty(object):
    """Decorator that adds class properties. Allows for usage like @property
    but applies the property at the class level. Helps avoid having to use
    metaclasses or other complex techniques to achieve similar results.
    """
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def is_sequence(obj):
    """Test if `obj` is an iterable but not dict or string. Mainly used to
    determine if `obj` can be treated like a list for iteration purposes.
    """
    return (isinstance(obj, Iterable)
            and not isinstance(obj, string_types)
            and not isinstance(obj, dict))


def has_primary_key(metadict):
    """Check if meta class' dict object has a primary key defined."""
    return any(column.primary_key
               for attr, column in iteritems(metadict)
               if isinstance(column, Column))


def base_columns_from_subquery(subquery):
    """Return non-aliased, base columns from subquery."""
    # base_columns is a set so we need to cast to list.
    return [(column, list(column.base_columns))
            for column in subquery.c.values()]


def join_subquery_on_columns(subquery, columns):
    """Return join-on condition which maps subquery's columns to columns."""
    subquery_base_columns = base_columns_from_subquery(subquery)

    join_on = []
    for subquery_column, base_columns in subquery_base_columns:
        # Don't support joining to subquery column with more than 1 base
        # column.
        if len(base_columns) == 1 and base_columns[0] in columns:
            join_on.append(subquery_column == base_columns[0])

    if join_on:
        return and_(*join_on)
    else:  # pragma: no cover
        return None


def camelcase_to_underscore(string):
    """Convert string from CamelCase to under_score"""
    regex_first_cap = re.compile('(.)([A-Z][a-z]+)')
    regex_all_cap = re.compile('([a-z0-9])([A-Z])')

    first_cap = regex_first_cap.sub(r'\1_\2', string)
    return regex_all_cap.sub(r'\1_\2', first_cap).lower()


def iterflatten(items):
    """Return iterator which flattens list/tuple of lists/tuples
    >>> to_flatten = [1, [2,3], [4, [5, [6]], 7], 8]
    >>> assert list(iterflatten(to_flatten)) == [1,2,3,4,5,6,7,8]
    """
    for item in items:
        if isinstance(item, (list, tuple)):
            for itm in flatten(item):
                yield itm
        else:
            yield item


def flatten(items):
    """Return flattened list of a list/tuple of lists/tuples
    >>> assert flatten([1, [2,3], [4, [5, [6]], 7], 8]) == [1,2,3,4,5,6,7,8]
    """
    return list(iterflatten(items))



"""Session class that supports binds.
"""

from sqlalchemy.orm.session import Session as SessionBase


class Session(SessionBase):
    """The session is the default session that Manager uses.
    It extends the default session system with bind selection.
    """

    def __init__(self, manager, **options):
        self.manager = manager
        bind = options.pop('bind', manager.engine)
        super(Session, self).__init__(
            bind=bind, binds=manager.binds_map, **options)

    def get_bind(self, mapper, clause=None):
        """Return engine bind using mapper info's bind_key if present."""
        # mapper is None if someone tries to just get a connection
        if mapper is not None:
            info = getattr(mapper.mapped_table, 'info', {})
            bind_key = info.get('bind_key')
            if bind_key is not None:
                return self.manager.get_engine(bind=bind_key)
        return super(Session, self).get_bind(mapper, clause)



# pylint: disable=C0103
"""SQLAlchemy query filter factories usable in
alchy.QueryModel.__search_filter__.

These are factory functions that return common filter operations as functions
which are then assigned to the model class' search config attributes.

For example:

    class UserQuery(QueryModel):
        @property
        def __search_filters__(self):
            return {
                'email': like(User.email)
            }

    class User(Model):
        query_class = UserQuery
        email = Column(types.String(100))

without using the search function:

    class UserQuery(QueryModel):
        __search_filters = {
            'email': lambda value: User.email.like(value)
        }

    class User(Model):
        query_class = UserQuery
        email = Column(types.String(100))

The general naming convention for each comparator is:

- positive comparator: base (e.g. "like")
- negative comparator: notbase (e.g. "notlike")
"""

from sqlalchemy import not_


class ColumnOperator(object):
    """Base class for column operator based search factories."""
    op = None

    def __init__(self, column):
        self.column = column

    def compare(self, value):
        """Return comparision with value."""
        return getattr(self.column, self.op)(value)

    def __call__(self, value):
        """Makes comparision."""
        if hasattr(self.column, '__call__'):
            self.column = self.column()
        return self.compare(value)


class NegateOperator(ColumnOperator):
    """Negates a ColumnOperator class."""
    def __call__(self, value):
        return not_(super(NegateOperator, self).__call__(value))


class RelationshipOperator(ColumnOperator):
    """Base class for relationship operator based search factories."""
    def __init__(self, column, column_operator):
        self.column = column
        self.column_operator = column_operator

    def compare(self, value):
        """Return comparision with value."""
        return getattr(self.column, self.op)(self.column_operator(value))


class like(ColumnOperator):
    """Return like filter function using ORM column field."""
    op = 'like'


class notlike(like, NegateOperator):
    """Return not like filter function using ORM column field."""
    pass


class ilike(ColumnOperator):
    """Return ilike filter function using ORM column field."""
    op = 'ilike'


class notilike(ilike, NegateOperator):
    """Return not ilike filter function using ORM column field."""
    pass


class startswith(ColumnOperator):
    """Return startswith filter function using ORM column field."""
    op = 'startswith'


class notstartswith(startswith, NegateOperator):
    """Return not startswith filter function using ORM column field."""
    pass


class endswith(ColumnOperator):
    """Return endswith filter function using ORM column field."""
    op = 'endswith'


class notendswith(endswith, NegateOperator):
    """Return not endswith filter function using ORM column field."""
    pass


class contains(ColumnOperator):
    """Return contains filter function using ORM column field."""
    op = 'contains'


class notcontains(contains, NegateOperator):
    """Return not contains filter function using ORM column field."""
    pass


class in_(ColumnOperator):
    """Return in_ filter function using ORM column field."""
    op = 'in_'


class notin_(in_, NegateOperator):
    """Return not in_ filter function using ORM column field."""
    pass


class eq(ColumnOperator):
    """Return == filter function using ORM column field."""
    def compare(self, value):
        return self.column == value


class noteq(eq, NegateOperator):
    """Return not(==) filter function using ORM column field."""
    pass


class gt(ColumnOperator):
    """Return > filter function using ORM column field."""
    def compare(self, value):
        return self.column > value


class notgt(gt, NegateOperator):
    """Return not(>) filter function using ORM column field."""
    pass


class ge(ColumnOperator):
    """Return >= filter function using ORM column field."""
    def compare(self, value):
        return self.column >= value


class notge(ge, NegateOperator):
    """Return not(>=) filter function using ORM column field."""
    pass


class lt(ColumnOperator):
    """Return < filter function using ORM column field."""
    def compare(self, value):
        return self.column < value


class notlt(lt, NegateOperator):
    """Return not(<) filter function using ORM column field."""
    pass


class le(ColumnOperator):
    """Return <= filter function using ORM column field."""
    def compare(self, value):
        return self.column <= value


class notle(le, NegateOperator):
    """Return not(<=) filter function using ORM column field."""
    pass


class any_(RelationshipOperator):
    """Return any filter function using ORM relationship field."""
    op = 'any'


class notany_(any_, NegateOperator):
    """Return not(any) filter function using ORM relationship field."""
    pass


class has(RelationshipOperator):
    """Return has filter function using ORM relationship field."""
    op = 'has'


class nothas(has, NegateOperator):
    """Return not(has) filter function using ORM relationship field."""
    pass



"""Manager class and mixin.

Provides `Manager` class for interacting with a database session via
SQLAlchemy's orm.scoped_session.
"""

from functools import partial

import sqlalchemy
from sqlalchemy import orm
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm.exc import UnmappedError

from .model import make_declarative_base, extend_declarative_base
from .query import QueryModel
from .session import Session
from ._compat import string_types, itervalues


class ManagerMixin(object):
    """Useful extensions to self.session."""

    def add(self, *instances):
        """Override `session.add()` so it can function like `session.add_all()`
        and support chaining.
        """
        for instance in instances:
            if isinstance(instance, list):
                self.add(*instance)
            else:
                self.session.add(instance)
        return self.session

    def add_commit(self, *instances):
        """Add instances to session and commit in one call."""
        self.add(*instances).commit()

    def delete(self, *instances):
        """Override `session.delete()` so it can function like
        `session.add_all()` and support chaining.
        """
        for instance in instances:
            if isinstance(instance, list):
                self.delete(*instance)
            else:
                self.session.delete(instance)
        return self.session

    def delete_commit(self, *instances):
        """Delete instances to session and commit in one call."""
        self.delete(*instances).commit()


class Manager(ManagerMixin):
    """Manager for session."""

    def __init__(self, config=None, session_options=None, Model=None):

        self.config = Config(defaults={
            'SQLALCHEMY_DATABASE_URI': 'sqlite://',
            'SQLALCHEMY_BINDS': None,
            'SQLALCHEMY_ECHO': False,
            'SQLALCHEMY_POOL_SIZE': None,
            'SQLALCHEMY_POOL_TIMEOUT': None,
            'SQLALCHEMY_POOL_RECYCLE': None,
            'SQLALCHEMY_MAX_OVERFLOW': None
        })

        if isinstance(config, dict):
            self.config.update(config)
        elif config is not None:
            self.config.from_object(config)

        self._engines = {}
        self._binds = {}

        if session_options is None:
            session_options = {}

        session_options.setdefault('query_cls', QueryModel)
        session_options.setdefault('autocommit', False)
        session_options.setdefault('autoflush', True)

        self.session = self.create_scoped_session(session_options)

        if Model is None:
            self.Model = make_declarative_base()
        else:
            self.Model = Model

        if self.Model:
            extend_declarative_base(self.Model, self.session)

    @property
    def metadata(self):
        """Return `Model`'s metadata object."""
        return getattr(self.Model, 'metadata', None)

    @property
    def binds(self):
        """Returns config options for all binds."""
        if not self._binds:
            self._binds = {
                None: self.config['SQLALCHEMY_DATABASE_URI']
            }

            if self.config['SQLALCHEMY_BINDS']:
                self._binds.update(self.config['SQLALCHEMY_BINDS'])

        return self._binds

    @property
    def binds_map(self):
        """Returns a dictionary with a table->engine mapping.
        This is suitable for use of sessionmaker(binds=binds_map).
        """
        binds = list(self.binds)
        retval = {}
        for bind in binds:
            engine = self.get_engine(bind)
            tables = self.get_tables_for_bind(bind)
            retval.update(dict((table, engine) for table in tables))
        return retval

    @property
    def engine(self):
        """Return default database engine."""
        return self.get_engine()

    def create_engine(self, uri_or_config):
        """Create engine using either a URI or a config dict.
        If URI supplied, then the default `self.config` will be used.
        If config supplied, then URI in config will be used.
        """
        if isinstance(uri_or_config, dict):
            uri = uri_or_config['SQLALCHEMY_DATABASE_URI']
            config = uri_or_config
        else:
            uri = uri_or_config
            config = self.config

        options = engine_options_from_config(config)

        return sqlalchemy.create_engine(make_url(uri), **options)

    def get_engine(self, bind=None):
        """Return engine associated with bind. Create engine if it
        doesn't already exist.
        """
        if bind not in self._engines:
            assert bind in self.binds, \
                'Bind {0} is not specified. ' \
                'Set in SQLALCHEMY_BINDS configuration variable'.format(bind)

            self._engines[bind] = self.create_engine(self.binds[bind])

        return self._engines[bind]

    def create_scoped_session(self, options=None):
        """Create scoped session which internally calls
        `self.create_session`.
        """
        if options is None:  # pragma: no cover
            options = {}
        return orm.scoped_session(partial(self.create_session, options))

    def create_session(self, options):
        """Create session instance using custom Session class that supports
        multiple bindings.
        """
        return Session(self, **options)

    def get_tables_for_bind(self, bind=None):
        """Returns a list of all tables relevant for a bind."""
        return [table
                for table in itervalues(self.metadata.tables)
                if table.info.get('bind_key') == bind]

    def _execute_for_all_tables(self, bind, operation, skip_tables=False):
        """Execute metadata operation for associated tables."""
        if self.metadata is None:
            raise UnmappedError('Missing declarative base model')

        if bind == '__all__':
            binds = [None] + list(self.config.get('SQLALCHEMY_BINDS') or {})
        elif isinstance(bind, string_types) or bind is None:
            binds = [bind]
        else:
            binds = bind

        for bind in binds:
            extra = {}
            if not skip_tables:
                tables = self.get_tables_for_bind(bind)
                extra['tables'] = tables
            metadata_operation = getattr(self.metadata, operation)
            metadata_operation(bind=self.get_engine(bind), **extra)

    def create_all(self, bind='__all__'):
        """Create database schema from models."""
        self._execute_for_all_tables(bind, 'create_all')

    def drop_all(self, bind='__all__'):
        """Drop tables defined by models."""
        self._execute_for_all_tables(bind, 'drop_all')

    def reflect(self, bind='__all__'):
        """Reflect tables from database."""
        self._execute_for_all_tables(bind, 'reflect', skip_tables=True)

    def __getattr__(self, attr):
        """Delegate all other attributes to self.session."""
        return getattr(self.session, attr)


class Config(dict):
    """Configuration loader which acts like a dict but supports loading
    values from an object limited to `ALL_CAPS_ATTRIBUTES`.
    """
    def __init__(self, defaults=None):
        super(Config, self).__init__(defaults or {})

    def from_object(self, obj):
        """Pull `dir(obj)` keys from and set onto self."""
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)


def engine_options_from_config(config):
    """Return engine options derived from config object."""
    options = {}

    def _setdefault(optionkey, configkey):
        """Set options key if config key is not None."""
        if config.get(configkey) is not None:
            options[optionkey] = config[configkey]

    _setdefault('echo', 'SQLALCHEMY_ECHO')
    _setdefault('pool_size', 'SQLALCHEMY_POOL_SIZE')
    _setdefault('pool_timeout', 'SQLALCHEMY_POOL_TIMEOUT')
    _setdefault('pool_recycle', 'SQLALCHEMY_POOL_RECYCLE')
    _setdefault('max_overflow', 'SQLALCHEMY_MAX_OVERFLOW')

    return options



# pylint: skip-file
"""Python 2/3 compatibility

    Some py2/py3 compatibility support based on a stripped down
    version of six so we don't have to depend on a specific version
    of it.

    Borrowed from
    https://github.com/mitsuhiko/flask/blob/master/flask/_compat.py
"""

import sys


PY3 = sys.version_info[0] == 3
_identity = lambda x: x


if PY3:
    text_type = str
    string_types = (str,)
    integer_types = (int,)

    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())

    from io import StringIO

    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    implements_to_string = _identity
else:
    text_type = unicode
    string_types = (str, unicode)
    integer_types = (int, long)

    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()

    from cStringIO import StringIO

    exec('def reraise(tp, value, tb=None):\n raise tp, value, tb')

    def implements_to_string(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode('utf-8')
        return cls


def with_metaclass(meta, *bases):
    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instantiation that replaces
    # itself with the actual metaclass.  Because of internal type checks
    # we also need to make sure that we downgrade the custom metaclass
    # for one level to something closer to type (that's why __call__ and
    # __init__ comes back from type etc.).
    #
    # This has the advantage over six.with_metaclass in that it does not
    # introduce dummy classes into the final MRO.
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})


# Certain versions of pypy have a bug where clearing the exception stack
# breaks the __exit__ function in a very peculiar way.  This is currently
# true for pypy 2.2.1 for instance.  The second level of exception blocks
# is necessary because pypy seems to forget to check if an exception
# happend until the next bytecode instruction?
BROKEN_PYPY_CTXMGR_EXIT = False
if hasattr(sys, 'pypy_version_info'):
    class _Mgr(object):
        def __enter__(self):
            return self

        def __exit__(self, *args):
            sys.exc_clear()
    try:
        try:
            with _Mgr():
                raise AssertionError()
        except:
            raise
    except TypeError:
        BROKEN_PYPY_CTXMGR_EXIT = True
    except AssertionError:
        pass



"""Query subclass used by Manager as default session query class.
"""

from math import ceil

from sqlalchemy import orm, and_, or_, inspect

try:
    from sqlalchemy.orm.strategy_options import Load
except ImportError:  # pragma: no cover
    import warnings
    from sqlalchemy import __version__ as sqla_version

    warn_msg = ('The installed version of SQLAlchemy=={0} is not compatible '
                'with the alchy.Query loading API. '
                'Use with caution!').format(sqla_version)

    warnings.warn(warn_msg, RuntimeWarning)

    class Load(object):
        """Dummy Load class which just fails when used."""
        def __call__(self, *args, **kargs):
            raise NotImplementedError(warn_msg)

        def __getattr__(self, attr):
            self()

from ._compat import iteritems
from .utils import join_subquery_on_columns


class Query(orm.Query):
    """Extension of default Query class used in SQLAlchemy session queries.
    """

    # Default per_page argument for pagination when per_page not specified.
    DEFAULT_PER_PAGE = 50

    @property
    def entities(self):
        """Return list of entity classes present in query"""
        return [e.mapper.class_ for e in self._entities]

    @property
    def join_entities(self):
        """Return list of the joined entity classes present in query"""
        return [e.mapper.class_ for e in self._join_entities]

    @property
    def all_entities(self):
        """Return list of entities + join_entities present in query"""
        return self.entities + self.join_entities

    def _join_eager(self, keys, outerjoin, **kargs):
        """Helper method for applying join()/outerjoin() with contains_eager()
        """
        alias = kargs.pop('alias', None)

        key = keys[0]
        path_keys = keys[1:]

        join_args = ([alias, key] if alias else [key]) + list(path_keys)

        opt = orm.contains_eager(key, alias=alias)

        for k in path_keys:
            opt = opt.contains_eager(k)

        join = self.outerjoin if outerjoin else self.join

        return join(*join_args).options(opt)

    def join_eager(self, *keys, **kargs):
        """Apply `self.join()` + `self.options(contains_eager())`."""
        return self._join_eager(keys, False, **kargs)

    def outerjoin_eager(self, *keys, **kargs):
        """Apply `self.outerjoin()` + `self.options(contains_eager())`."""
        return self._join_eager(keys, True, **kargs)

    def _join_load(self, keys, load_type, **kargs):
        """Helper method for returning load strategies."""
        opt = getattr(orm, load_type)(keys[0], **kargs)

        for k in keys[1:]:
            opt = getattr(opt, load_type)(k)

        return self.options(opt)

    def joinedload(self, *keys, **kargs):
        """Apply joinedload() to `keys`"""
        return self._join_load(keys, 'joinedload', **kargs)

    def immediateload(self, *keys, **kargs):
        """Apply immediateload() to `keys`"""
        return self._join_load(keys, 'immediateload', **kargs)

    def lazyload(self, *keys, **kargs):
        """Apply lazyload() to `keys`"""
        return self._join_load(keys, 'lazyload', **kargs)

    def noload(self, *keys, **kargs):
        """Apply noload() to `keys`"""
        return self._join_load(keys, 'noload', **kargs)

    def subqueryload(self, *keys, **kargs):
        """Apply subqueryload() to `keys`"""
        return self._join_load(keys, 'subqueryload', **kargs)

    def load_only(self, *columns):
        """Apply `load_only()` to query"""
        obj, columns = get_load_options(*columns)
        return self.options(obj.load_only(*columns))

    def defer(self, *columns):
        """Apply `defer()` to query"""
        obj, columns = get_load_options(*columns)
        opts = obj
        for column in columns:
            opts = opts.defer(column)
        return self.options(opts)

    def undefer(self, *columns):
        """Apply `undefer()` to query"""
        obj, columns = get_load_options(*columns)
        opts = obj
        for column in columns:
            opts = opts.undefer(column)
        return self.options(opts)

    def undefer_group(self, *names):
        """Apply `undefer_group()` to query"""
        obj, names = get_load_options(*names)
        return self.options(obj.undefer_group(names[0]))

    def map(self, func):
        """Call native `map` on `self.all()`"""
        return [func(item) for item in self.all()]

    def reduce(self, func, initial=None, reverse=False):
        """Reduce `self.all()` using `func`"""
        items = self.all()

        if reverse:
            items = reversed(items)

        result = initial
        for item in items:
            result = func(result, item)
        return result

    def reduce_right(self, func, initial=None):
        """Reduce reversed `self.all()` using `func`"""
        return self.reduce(func, initial=initial, reverse=True)

    def pluck(self, column):
        """Pluck `column` attribute values from `self.all()` results and return
        as list.
        """
        return [getattr(r, column, None) for r in self.all()]

    def page(self, page=1, per_page=None):
        """Return query with limit and offset applied for page"""
        if per_page is None:
            per_page = self.DEFAULT_PER_PAGE

        return self.limit(per_page).offset((page - 1) * per_page)

    def paginate(self, page=1, per_page=None, error_out=True):
        """Return Pagination instance using already defined query parameters"""
        if error_out and page < 1:
            raise IndexError

        if per_page is None:
            per_page = self.DEFAULT_PER_PAGE

        items = self.page(page, per_page).all()

        if not items and page != 1 and error_out:
            raise IndexError

        # No need to count if we're on the first page and there are fewer items
        # than we expected.
        if page == 1 and len(items) < per_page:
            total = len(items)
        else:
            total = self.order_by(None).count()

        return Pagination(self, page, per_page, total, items)


class QueryModel(Query):
    """Class used for default query property class. It's assumed that is how
    this class will be used so there may be cases where some of its functions
    won't work properly outside that context.
    """

    # All available search filter functions indexed by a canonical name which
    # will be referenced in advanced/simple search. All filter functions should
    # take a single value and return an SQLAlchemy filter expression.
    # I.e. {key: lambda value: Model.column_name.contains(value)}
    __search_filters__ = {}

    # Advanced search models search by named parameters. Generally found on
    # advanced search forms where each field maps to a specific database field
    # that will be queried against. If defined as a list, each item should be
    # a key from __search_filters__. The matching __search_filters__ function
    # will be used in the query. If defined as a dict, it should have the same
    # format as __search_filters__.
    __advanced_search__ = []

    # Simple search models search by phrase (like Google search). Defined like
    # __advanced_search__.
    __simple_search__ = []

    @property
    def Model(self):
        """Return primary entity model class."""
        return self.entities[0]

    def get_search_filters(self, keys):
        """Return __search_filters__ filtered by keys."""
        if isinstance(keys, dict):
            return keys
        else:
            return dict([(key, self.__search_filters__[key]) for key in keys])

    def advanced_filter(self, search_dict=None):
        """Return the compiled advanced search filter mapped to search_dict."""
        if search_dict is None:  # pragma: no cover
            search_dict = {}

        filter_funcs = self.get_search_filters(self.__advanced_search__)
        term_filters = [filter_funcs[key](value)
                        for key, value in iteritems(search_dict)
                        if key in filter_funcs]

        # All filters should match for an advanced search.
        return and_(*term_filters)

    def simple_filter(self, search_terms=None):
        """Return the compiled simple search filter mapped to search_terms."""
        if search_terms is None:  # pragma: no cover
            search_terms = []

        filter_funcs = self.get_search_filters(self.__simple_search__)

        # Only support AND'ing search terms together. Apply each simple search
        # filter to each search term and group them together.
        term_filters = [[func(term) for func in filter_funcs.values()]
                        for term in search_terms]

        # Each item in term_filters is a list of filters applied to one of
        # the search terms contained in search_string. We need at least one
        # simple filter to match for each term. We need all search terms to
        # have at least simple filter match.
        return and_(*[or_(*filters) for filters in term_filters])

    def search(self, search_string=None, search_dict=None,
               limit=None, offset=None, order_by=None):
        """Perform combination of simple/advanced searching with optional
        limit/offset support.
        """
        model_primary_keys = self.Model.primary_keys

        # Apply search filtering and pagination to Model's primary keys so we
        # can use the query as a subquery. In order to properly handle
        # pagination, we should use a subquery so that the outer level joins
        # won't cause records to be excluded when those includes *-to-many
        # relationships. For example, if we were returning a query of user +
        # user keywords (one-to-many), then for something like the first 25
        # users, we may actually have more than that many records since we're
        # joining on many records from the user keywords table.
        model_query = original = self.session.query(
            *model_primary_keys).distinct()

        if search_string is not None:
            model_query = model_query.filter(
                self.simple_filter(search_string.split()))

        if search_dict is not None:
            model_query = model_query.filter(
                self.advanced_filter(search_dict))

        if limit is not None:
            model_query = model_query.limit(limit)

        if offset is not None:
            model_query = model_query.offset(offset)

        if order_by is not None:
            model_query = model_query.order_by(order_by)

        if model_query != original:
            subquery = model_query.subquery()

            query = self.join(subquery, join_subquery_on_columns(
                subquery, model_primary_keys))
        else:
            query = self

        return query


class QueryProperty(object):
    """Query property accessor which gives a model access to query capabilities
    via Model.query which is equivalent to session.query(Model)
    """
    def __init__(self, session):
        self.session = session

    def __get__(self, model, Model):
        mapper = orm.class_mapper(Model)
        if mapper:
            if not getattr(Model, 'query_class', None):
                Model.query_class = QueryModel

            query_property = Model.query_class(mapper, session=self.session())

            return query_property


##
# Pagination class and usage adapated from Flask-SQLAlchemy:
# https://github.com/mitsuhiko/flask-sqlalchemy
##


class Pagination(object):
    """
    Internal helper class returned by `BaseQuery.paginate`.  You
    can also construct it from any other SQLAlchemy query object if you are
    working with other libraries.  Additionally it is possible to pass `None`
    as query object in which case the `prev` and `next` will
    no longer work.
    """

    def __init__(self, query, page, per_page, total, items):
        # the unlimited query object that was used to create this pagination
        # object
        self.query = query

        # the current page number (1 indexed)
        self.page = page

        # the number of items to be displayed on a page.
        self.per_page = per_page

        # the total number of items matching the query
        self.total = total

        # the items for the current page
        self.items = items

        # the total number of pages
        if self.per_page == 0:
            self.pages = 0
        else:
            self.pages = int(ceil(self.total / float(self.per_page)))

        # number of the previous page
        self.prev_num = self.page - 1

        # True if a previous page exists
        self.has_prev = self.page > 1

        # number of the next page
        self.next_num = self.page + 1

        # True if a next page exists
        self.has_next = self.page < self.pages

    def prev(self, error_out=False):
        """Returns a `Pagination` object for the previous page."""
        assert self.query is not None, \
            'a query object is required for this method to work'
        return self.query.paginate(self.page - 1, self.per_page, error_out)

    def next(self, error_out=False):
        """Returns a `Pagination` object for the next page."""
        assert self.query is not None, \
            'a query object is required for this method to work'
        return self.query.paginate(self.page + 1, self.per_page, error_out)


def get_load_options(*columns):
    """Helper method that attempts to extract a sqlalchemy object from
    `columns[0]` and return remaining columns to apply to a query load method.
    """
    model_inspect = inspect(columns[0], raiseerr=False)

    # return an obj which has loading API
    if model_inspect and model_inspect.is_mapper:
        obj = Load(columns[0])
        columns = columns[1:]
    elif isinstance(columns[0], Load):
        obj = columns[0]
        columns = columns[1:]
    else:
        obj = orm

    return (obj, columns)
